import time
import uuid
import zipfile

import requests

import dbt.tracking
from dbt._pydantic_shim import BaseSettings  # type: ignore
from dbt.config.runtime import UnsetProfile, load_project
from dbt.constants import MANIFEST_FILE_NAME, RUN_RESULTS_FILE_NAME
from dbt.events.types import ArtifactUploadSkipped, ArtifactUploadSuccess
from dbt.exceptions import DbtProjectError
from dbt_common.events.functions import fire_event
from dbt_common.exceptions import DbtBaseException as DbtException

MAX_RETRIES = 3

EXECUTION_ARTIFACTS = [MANIFEST_FILE_NAME, RUN_RESULTS_FILE_NAME]

PRODUCED_ARTIFACTS_PATHS: set[str] = set()


# artifact paths calling this will be uploaded to dbt Cloud
def add_artifact_produced(artifact_path: str):
    PRODUCED_ARTIFACTS_PATHS.add(artifact_path)


class ArtifactUploadConfig(BaseSettings):
    tenant_hostname: str
    DBT_CLOUD_TOKEN: str
    DBT_CLOUD_ACCOUNT_ID: str
    DBT_CLOUD_ENVIRONMENT_ID: str

    def get_ingest_url(self):
        return f"https://{self.tenant_hostname}/api/private/accounts/{self.DBT_CLOUD_ACCOUNT_ID}/environments/{self.DBT_CLOUD_ENVIRONMENT_ID}/ingests/"

    def get_complete_url(self, ingest_id):
        return f"{self.get_ingest_url()}{ingest_id}/"

    def get_headers(self, invocation_id=None):
        if invocation_id is None:
            invocation_id = str(uuid.uuid4())
        return {
            "Accept": "application/json",
            "X-Invocation-Id": invocation_id,
            "Authorization": f"Token {self.DBT_CLOUD_TOKEN}",
        }


def _retry_with_backoff(operation_name, func, max_retries=MAX_RETRIES, retry_codes=None):
    """Execute a function with exponential backoff retry logic.

    Args:
        operation_name: Name of the operation for error messages
        func: Function to execute that returns (success, result)
        max_retries: Maximum number of retry attempts

    Returns:
        The result from the function if successful

    Raises:
        DbtException: If all retry attempts fail
    """
    if retry_codes is None:
        retry_codes = [500, 502, 503, 504]
    retry_delay = 1
    for attempt in range(max_retries + 1):
        try:
            success, result = func()
            if success:
                return result

            if result.status_code not in retry_codes:
                raise DbtException(f"Error {operation_name}: {result}")
            if attempt == max_retries:  # Last attempt
                raise DbtException(f"Error {operation_name}: {result}")
        except requests.RequestException as e:
            if attempt == max_retries:  # Last attempt
                raise DbtException(f"Error {operation_name}: {str(e)}")

        time.sleep(retry_delay)
        retry_delay *= 2  # exponential backoff


def upload_artifacts(project_dir, target_path, command):
    # Check if there are artifacts to upload for this command
    if not PRODUCED_ARTIFACTS_PATHS:
        fire_event(ArtifactUploadSkipped(msg="No artifacts to upload for current command"))
        return

    # read configurations
    try:
        project = load_project(
            project_dir, version_check=False, profile=UnsetProfile(), cli_vars=None
        )
        if not project.dbt_cloud or "tenant_hostname" not in project.dbt_cloud:
            raise DbtProjectError("dbt_cloud.tenant_hostname not found in dbt_project.yml")
        tenant_hostname = project.dbt_cloud["tenant_hostname"]
        if not tenant_hostname:
            raise DbtProjectError("dbt_cloud.tenant_hostname is empty in dbt_project.yml")
    except Exception as e:
        raise DbtProjectError(
            f"Error reading dbt_cloud.tenant_hostname from dbt_project.yml: {str(e)}"
        )

    config = ArtifactUploadConfig(tenant_hostname=tenant_hostname)

    if not target_path:
        target_path = "target"

    # Create zip file with artifacts
    zip_file_name = "target.zip"
    with zipfile.ZipFile(zip_file_name, "w") as z:
        for artifact_path in PRODUCED_ARTIFACTS_PATHS:
            z.write(artifact_path, artifact_path.split("/")[-1])

    # Step 1: Create ingest request with retry
    def create_ingest():
        response = requests.post(url=config.get_ingest_url(), headers=config.get_headers())
        return response.status_code == 200, response

    response = _retry_with_backoff("creating ingest request", create_ingest)
    response_data = response.json()
    ingest_id = response_data["data"]["id"]
    upload_url = response_data["data"]["upload_url"]

    # Step 2: Upload the zip file to the provided URL with retry
    with open(zip_file_name, "rb") as f:
        file_data = f.read()

        def upload_file():
            upload_response = requests.put(url=upload_url, data=file_data)
            return upload_response.status_code in (200, 204), upload_response

        _retry_with_backoff("uploading artifacts", upload_file)

    # Step 3: Mark the ingest as successful with retry
    def complete_ingest():
        complete_response = requests.patch(
            url=config.get_complete_url(ingest_id),
            headers=config.get_headers(),
            json={"upload_status": "SUCCESS"},
        )
        return complete_response.status_code == 204, complete_response

    _retry_with_backoff("completing ingest", complete_ingest)

    fire_event(ArtifactUploadSuccess(msg=f"command {command} completed successfully"))
    if dbt.tracking.active_user is not None:
        dbt.tracking.track_artifact_upload({"command": command})
    PRODUCED_ARTIFACTS_PATHS.clear()



import os
from copy import deepcopy
from typing import Any, Dict, List, Optional

from dbt.artifacts.resources import Catalog, CatalogWriteIntegrationConfig
from dbt.clients.yaml_helper import load_yaml_text
from dbt.config.renderer import SecretRenderer
from dbt.constants import CATALOGS_FILE_NAME
from dbt.exceptions import YamlLoadError
from dbt_common.clients.system import load_file_contents
from dbt_common.exceptions import CompilationError, DbtValidationError


def load_catalogs_yml(project_dir: str, project_name: str) -> Dict[str, Any]:
    path = os.path.join(project_dir, CATALOGS_FILE_NAME)

    if os.path.isfile(path):
        try:
            contents = load_file_contents(path, strip=False)
            yaml_content = load_yaml_text(contents)

            if not yaml_content:
                raise DbtValidationError(f"The file at {path} is empty")

            return yaml_content
        except DbtValidationError as e:
            raise YamlLoadError(project_name=project_name, path=CATALOGS_FILE_NAME, exc=e)

    return {}


def load_single_catalog(raw_catalog: Dict[str, Any], renderer: SecretRenderer) -> Catalog:
    try:
        rendered_catalog = renderer.render_data(raw_catalog)
    except CompilationError as exc:
        raise DbtValidationError(str(exc)) from exc

    Catalog.validate(rendered_catalog)

    write_integrations = []
    write_integration_names = set()

    for raw_integration in rendered_catalog.get("write_integrations", []):
        if raw_integration["name"] in write_integration_names:
            raise DbtValidationError(
                f"Catalog '{rendered_catalog['name']}' cannot have multiple 'write_integrations' with the same name: '{raw_integration['name']}'."
            )

        # We're going to let the adapter validate the integration config
        write_integrations.append(
            CatalogWriteIntegrationConfig(**raw_integration, catalog_name=raw_catalog["name"])
        )
        write_integration_names.add(raw_integration["name"])

    # Validate + set default active_write_integration if unset
    active_write_integration = rendered_catalog.get("active_write_integration")
    valid_write_integration_names = [integration.name for integration in write_integrations]

    if not active_write_integration:
        if len(valid_write_integration_names) == 1:
            active_write_integration = write_integrations[0].name
        else:
            raise DbtValidationError(
                f"Catalog '{rendered_catalog['name']}' must specify an 'active_write_integration' when multiple 'write_integrations' are provided."
            )
    else:
        if active_write_integration not in valid_write_integration_names:
            raise DbtValidationError(
                f"Catalog '{rendered_catalog['name']}' must specify an 'active_write_integration' from its set of defined 'write_integrations': {valid_write_integration_names}. Got: '{active_write_integration}'."
            )

    return Catalog(
        name=raw_catalog["name"],
        active_write_integration=active_write_integration,
        write_integrations=write_integrations,
    )


def load_catalogs(project_dir: str, project_name: str, cli_vars: Dict[str, Any]) -> List[Catalog]:
    raw_catalogs = load_catalogs_yml(project_dir, project_name).get("catalogs", [])
    catalogs_renderer = SecretRenderer(cli_vars)

    return [load_single_catalog(raw_catalog, catalogs_renderer) for raw_catalog in raw_catalogs]


def get_active_write_integration(catalog: Catalog) -> Optional[CatalogWriteIntegrationConfig]:
    for integration in catalog.write_integrations:
        if integration.name == catalog.active_write_integration:
            active_integration = deepcopy(integration)
            active_integration.catalog_name = active_integration.name
            active_integration.name = catalog.name
            return active_integration

    return None



import json
import os
import re
from datetime import date, datetime
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Union

import jsonschema
from jsonschema import ValidationError
from jsonschema._keywords import type as type_rule
from jsonschema.validators import Draft7Validator, extend

from dbt import deprecations
from dbt.include.jsonschemas import JSONSCHEMAS_PATH
from dbt_common.context import get_invocation_context

_PROJECT_SCHEMA: Optional[Dict[str, Any]] = None
_RESOURCES_SCHEMA: Optional[Dict[str, Any]] = None

_JSONSCHEMA_SUPPORTED_ADAPTERS = {
    "bigquery",
    "databricks",
    "redshift",
    "snowflake",
}

_HIERARCHICAL_CONFIG_KEYS = {
    "seeds",
    "sources",
    "models",
    "snapshots",
    "tests",
    "exposures",
    "data_tests",
    "metrics",
    "saved_queries",
    "semantic_models",
    "unit_tests",
}


def load_json_from_package(jsonschema_type: str, filename: str) -> Dict[str, Any]:
    """Loads a JSON file from within a package."""

    path = Path(JSONSCHEMAS_PATH).joinpath(jsonschema_type, filename)
    data = path.read_bytes()
    return json.loads(data)


def project_schema() -> Dict[str, Any]:
    global _PROJECT_SCHEMA

    if _PROJECT_SCHEMA is None:
        _PROJECT_SCHEMA = load_json_from_package(
            jsonschema_type="project", filename="0.0.110.json"
        )
    return _PROJECT_SCHEMA


def resources_schema() -> Dict[str, Any]:
    global _RESOURCES_SCHEMA

    if _RESOURCES_SCHEMA is None:
        _RESOURCES_SCHEMA = load_json_from_package(
            jsonschema_type="resources", filename="latest.json"
        )

    return _RESOURCES_SCHEMA


def custom_type_rule(validator, types, instance, schema):
    """This is necessary because PyYAML loads things that look like dates or datetimes as those
    python objects. Then jsonschema.validate() fails because it expects strings.
    """
    if "string" in types and (isinstance(instance, datetime) or isinstance(instance, date)):
        return
    else:
        return type_rule(validator, types, instance, schema)


CustomDraft7Validator = extend(Draft7Validator, validators={"type": custom_type_rule})


def error_path_to_string(error: jsonschema.ValidationError) -> str:
    if len(error.path) == 0:
        return ""
    else:
        path = str(error.path.popleft())
        for part in error.path:
            if isinstance(part, int):
                path += f"[{part}]"
            else:
                path += f".{part}"

        return path


def _additional_properties_violation_keys(error: ValidationError) -> List[str]:
    found_keys = re.findall(r"'\S+'", error.message)
    return [key.strip("'") for key in found_keys]


def _validate_with_schema(
    schema: Dict[str, Any], json: Dict[str, Any]
) -> Iterator[ValidationError]:
    validator = CustomDraft7Validator(schema)
    return validator.iter_errors(json)


def _get_allowed_config_fields_from_error_path(
    yml_schema: Dict[str, Any], error_path: List[Union[str, int]]
) -> Optional[List[str]]:
    property_field_name = None
    node_schema = yml_schema["properties"]
    for part in error_path:
        if isinstance(part, str):
            if part in node_schema:
                if "items" not in node_schema[part]:
                    break

                # Update property field name
                property_field_name = node_schema[part]["items"]["$ref"].split("/")[-1]

                # Jump to the next level of the schema
                item_definition = node_schema[part]["items"]["$ref"].split("/")[-1]
                node_schema = yml_schema["definitions"][item_definition]["properties"]

    if not property_field_name:
        return None

    if "config" not in yml_schema["definitions"][property_field_name]["properties"]:
        return None

    config_field_name = yml_schema["definitions"][property_field_name]["properties"]["config"][
        "anyOf"
    ][0]["$ref"].split("/")[-1]

    allowed_config_fields = list(set(yml_schema["definitions"][config_field_name]["properties"]))

    return allowed_config_fields


def _can_run_validations() -> bool:
    if not os.environ.get("DBT_ENV_PRIVATE_RUN_JSONSCHEMA_VALIDATIONS"):
        return False

    invocation_context = get_invocation_context()
    return invocation_context.adapter_types.issubset(_JSONSCHEMA_SUPPORTED_ADAPTERS)


def jsonschema_validate(schema: Dict[str, Any], json: Dict[str, Any], file_path: str) -> None:
    if not _can_run_validations():
        return

    errors = _validate_with_schema(schema, json)
    for error in errors:
        # Listify the error path to make it easier to work with (it's a deque in the ValidationError object)
        error_path = list(error.path)
        if error.validator == "additionalProperties":
            keys = _additional_properties_violation_keys(error)
            if len(error.path) == 0:
                for key in keys:
                    deprecations.warn(
                        "custom-top-level-key-deprecation",
                        msg="Unexpected top-level key" + (" " + key if key else ""),
                        file=file_path,
                    )
            else:
                key_path = error_path_to_string(error)
                for key in keys:
                    if key == "overrides" and key_path.startswith("sources"):

                        deprecations.warn(
                            "source-override-deprecation",
                            source_name=key_path.split(".")[-1],
                            file=file_path,
                        )
                    else:
                        allowed_config_fields = _get_allowed_config_fields_from_error_path(
                            schema, error_path
                        )
                        if allowed_config_fields and key in allowed_config_fields:
                            deprecations.warn(
                                "property-moved-to-config-deprecation",
                                key=key,
                                file=file_path,
                                key_path=key_path,
                            )
                        else:
                            deprecations.warn(
                                "custom-key-in-object-deprecation",
                                key=key,
                                file=file_path,
                                key_path=key_path,
                            )
        elif error.validator == "anyOf" and len(error_path) > 0:
            sub_errors = error.context or []
            # schema yaml resource configs
            if error_path[-1] == "config":
                for sub_error in sub_errors:
                    if (
                        isinstance(sub_error, ValidationError)
                        and sub_error.validator == "additionalProperties"
                    ):
                        keys = _additional_properties_violation_keys(sub_error)
                        key_path = error_path_to_string(error)
                        for key in keys:
                            deprecations.warn(
                                "custom-key-in-config-deprecation",
                                key=key,
                                file=file_path,
                                key_path=key_path,
                            )
            # dbt_project.yml configs
            elif "dbt_project.yml" in file_path and error_path[0] in _HIERARCHICAL_CONFIG_KEYS:
                for sub_error in sub_errors:
                    if isinstance(sub_error, ValidationError) and sub_error.validator == "type":
                        # Only raise type-errors if they are indicating leaf config without a plus prefix
                        if (
                            len(sub_error.path) > 0
                            and isinstance(sub_error.path[-1], str)
                            and not sub_error.path[-1].startswith("+")
                        ):
                            deprecations.warn(
                                "missing-plus-prefix-in-config-deprecation",
                                key=sub_error.path[-1],
                                file=file_path,
                                key_path=error_path_to_string(sub_error),
                            )
        elif error.validator == "type":
            # Not deprecating invalid types yet
            pass
        else:
            deprecations.warn(
                "generic-json-schema-validation-deprecation",
                violation=error.message,
                file=file_path,
                key_path=error_path_to_string(error),
            )


def validate_model_config(config: Dict[str, Any], file_path: str) -> None:
    if not _can_run_validations():
        return

    resources_jsonschema = resources_schema()
    nested_definition_name = "ModelConfig"

    model_config_schema = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "title": nested_definition_name,
        **resources_jsonschema["definitions"][nested_definition_name],
        "definitions": {
            k: v
            for k, v in resources_jsonschema["definitions"].items()
            if k != nested_definition_name
        },
    }

    errors = _validate_with_schema(model_config_schema, config)
    for error in errors:
        error_path = list(error.path)
        if error.validator == "additionalProperties":
            keys = _additional_properties_violation_keys(error)
            if len(error.path) == 0:
                key_path = error_path_to_string(error)
                for key in keys:
                    deprecations.warn(
                        "custom-key-in-config-deprecation",
                        key=key,
                        file=file_path,
                        key_path=key_path,
                    )
            else:
                error.path.appendleft("config")
                key_path = error_path_to_string(error)
                for key in keys:
                    deprecations.warn(
                        "custom-key-in-object-deprecation",
                        key=key,
                        file=file_path,
                        key_path=key_path,
                    )
        elif error.validator == "type":
            # Not deprecating invalid types yet, except for pre-existing deprecation_date deprecation
            pass
        elif error.validator == "anyOf" and len(error_path) > 0:
            for sub_error in error.context or []:
                if (
                    isinstance(sub_error, ValidationError)
                    and sub_error.validator == "additionalProperties"
                ):
                    error.path.appendleft("config")
                    keys = _additional_properties_violation_keys(sub_error)
                    key_path = error_path_to_string(error)
                    for key in keys:
                        deprecations.warn(
                            "custom-key-in-object-deprecation",
                            key=key,
                            file=file_path,
                            key_path=key_path,
                        )
        else:
            deprecations.warn(
                "generic-json-schema-validation-deprecation",
                violation=error.message,
                file=file_path,
                key_path=error_path_to_string(error),
            )



from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from dbt.adapters.catalogs import CatalogIntegrationConfig
from dbt_common.dataclass_schema import dbtClassMixin


@dataclass
class CatalogWriteIntegrationConfig(CatalogIntegrationConfig):
    name: str
    catalog_type: str
    external_volume: Optional[str] = None
    table_format: Optional[str] = None
    catalog_name: Optional[str] = None
    file_format: Optional[str] = None
    adapter_properties: Dict[str, Any] = field(default_factory=dict)


@dataclass
class Catalog(dbtClassMixin):
    name: str
    active_write_integration: Optional[str] = None
    write_integrations: List[CatalogWriteIntegrationConfig] = field(default_factory=list)



import collections
import dataclasses
from typing import Any, Dict, List, Optional, Tuple

import yaml

from dbt import deprecations
from dbt.clients.yaml_helper import load_yaml_text

# the C version is faster, but it doesn't always exist
try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    from yaml import SafeLoader  # type: ignore  # noqa: F401


@dataclasses.dataclass
class YamlCheckFailure:
    failure_type: str
    message: str


def checked_load(contents) -> Tuple[Optional[Dict[str, Any]], List[YamlCheckFailure]]:
    # A hacky (but sadly justified) method for modifying a bit of PyYAML. We create
    # a new local subclass of SafeLoader, since we need to associate state with
    # the static class, but static classes do not have non-static state. This allows
    # us to be sure we have exclusive access to the class.
    class CheckedLoader(SafeLoader):
        check_failures: List[YamlCheckFailure] = []

        def construct_mapping(self, node, deep=False):
            if not isinstance(node, yaml.MappingNode):
                raise yaml.constructor.ConstructorError(
                    None, None, "expected a mapping node, but found %s" % node.id, node.start_mark
                )
            is_override = (
                len(node.value) > 0
                and len(node.value[0]) > 0
                and getattr(node.value[0][0], "value") == "<<"
            )
            self.flatten_mapping(node)
            mapping = {}
            for key_node, value_node in node.value:
                key = self.construct_object(key_node, deep=deep)
                if not isinstance(key, collections.abc.Hashable):
                    raise yaml.constructor.ConstructorError(
                        "while constructing a mapping",
                        node.start_mark,
                        "found unhashable key",
                        key_node.start_mark,
                    )
                value = self.construct_object(value_node, deep=deep)

                if not is_override and key in mapping:
                    start_mark = str(key_node.start_mark)
                    if start_mark.startswith("  in"):  # this means it was at the top level
                        message = f"Duplicate key '{key}' {start_mark.lstrip()}"
                    else:
                        message = f"Duplicate key '{key}' at {key_node.start_mark}"

                    self.check_failures.append(YamlCheckFailure("duplicate_key", message))

                mapping[key] = value
            return mapping

    CheckedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, CheckedLoader.construct_mapping
    )

    dct = load_yaml_text(contents, loader=CheckedLoader)
    check_failures = CheckedLoader.check_failures

    return (dct, check_failures)


def issue_deprecation_warnings_for_failures(failures: List[YamlCheckFailure], file: str):
    for failure in failures:
        if failure.failure_type == "duplicate_key":
            deprecations.warn(
                "duplicate-yaml-keys-deprecation",
                duplicate_description=failure.message,
                file=file,
            )



import abc
from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Callable, ClassVar, DefaultDict, Dict, List, Optional

import dbt.tracking
from dbt.events import types as core_types
from dbt.flags import get_flags
from dbt_common.dataclass_schema import dbtClassMixin
from dbt_common.events.functions import fire_event, warn_or_error
from dbt_common.events.types import Note


class DBTDeprecation:
    _name: ClassVar[Optional[str]] = None
    _event: ClassVar[Optional[str]] = None
    _is_preview: ClassVar[bool] = False

    @property
    def name(self) -> str:
        if self._name is not None:
            return self._name
        raise NotImplementedError("name not implemented for {}".format(self))

    def track_deprecation_warn(self) -> None:
        if dbt.tracking.active_user is not None:
            dbt.tracking.track_deprecation_warn({"deprecation_name": self.name})

    @property
    def event(self) -> abc.ABCMeta:
        if self._event is not None:
            module_path = core_types
            class_name = self._event

            try:
                return getattr(module_path, class_name)
            except AttributeError:
                msg = f"Event Class `{class_name}` is not defined in `{module_path}`"
                raise NameError(msg)
        raise NotImplementedError("event not implemented for {}".format(self._event))

    def preview(self, base_event: abc.ABCMeta) -> None:
        note_event = Note(msg=base_event.message())  # type: ignore
        fire_event(note_event)

    def show(self, *args, **kwargs) -> None:
        if self._is_preview:
            base_event = self.event(**kwargs)
            self.preview(base_event)
        else:
            flags = get_flags()
            if self.name not in active_deprecations or flags.show_all_deprecations:
                event = self.event(**kwargs)
                warn_or_error(event)
                self.track_deprecation_warn()

            active_deprecations[self.name] += 1


class PackageRedirectDeprecation(DBTDeprecation):
    _name = "package-redirect"
    _event = "PackageRedirectDeprecation"


class PackageInstallPathDeprecation(DBTDeprecation):
    _name = "install-packages-path"
    _event = "PackageInstallPathDeprecation"


# deprecations with a pattern of `project-config-*` for the name are not hardcoded
# they are called programatically via the pattern below
class ConfigSourcePathDeprecation(DBTDeprecation):
    _name = "project-config-source-paths"
    _event = "ConfigSourcePathDeprecation"


class ConfigDataPathDeprecation(DBTDeprecation):
    _name = "project-config-data-paths"
    _event = "ConfigDataPathDeprecation"


class ConfigLogPathDeprecation(DBTDeprecation):
    _name = "project-config-log-path"
    _event = "ConfigLogPathDeprecation"


class ConfigTargetPathDeprecation(DBTDeprecation):
    _name = "project-config-target-path"
    _event = "ConfigTargetPathDeprecation"


def renamed_method(old_name: str, new_name: str):
    class AdapterDeprecationWarning(DBTDeprecation):
        _name = "adapter:{}".format(old_name)
        _event = "AdapterDeprecationWarning"

    dep = AdapterDeprecationWarning()
    deprecations_list.append(dep)
    deprecations[dep.name] = dep


class MetricAttributesRenamed(DBTDeprecation):
    _name = "metric-attr-renamed"
    _event = "MetricAttributesRenamed"


class ExposureNameDeprecation(DBTDeprecation):
    _name = "exposure-name"
    _event = "ExposureNameDeprecation"


class CollectFreshnessReturnSignature(DBTDeprecation):
    _name = "collect-freshness-return-signature"
    _event = "CollectFreshnessReturnSignature"


class ProjectFlagsMovedDeprecation(DBTDeprecation):
    _name = "project-flags-moved"
    _event = "ProjectFlagsMovedDeprecation"


class PackageMaterializationOverrideDeprecation(DBTDeprecation):
    _name = "package-materialization-override"
    _event = "PackageMaterializationOverrideDeprecation"


class ResourceNamesWithSpacesDeprecation(DBTDeprecation):
    _name = "resource-names-with-spaces"
    _event = "ResourceNamesWithSpacesDeprecation"


class SourceFreshnessProjectHooksNotRun(DBTDeprecation):
    _name = "source-freshness-project-hooks"
    _event = "SourceFreshnessProjectHooksNotRun"


class MFTimespineWithoutYamlConfigurationDeprecation(DBTDeprecation):
    _name = "mf-timespine-without-yaml-configuration"
    _event = "MFTimespineWithoutYamlConfigurationDeprecation"


class MFCumulativeTypeParamsDeprecation(DBTDeprecation):
    _name = "mf-cumulative-type-params-deprecation"
    _event = "MFCumulativeTypeParamsDeprecation"


class MicrobatchMacroOutsideOfBatchesDeprecation(DBTDeprecation):
    _name = "microbatch-macro-outside-of-batches-deprecation"
    _event = "MicrobatchMacroOutsideOfBatchesDeprecation"


class GenericJSONSchemaValidationDeprecation(DBTDeprecation):
    _name = "generic-json-schema-validation-deprecation"
    _event = "GenericJSONSchemaValidationDeprecation"
    _is_preview = True


class UnexpectedJinjaBlockDeprecation(DBTDeprecation):
    _name = "unexpected-jinja-block-deprecation"
    _event = "UnexpectedJinjaBlockDeprecation"


class DuplicateYAMLKeysDeprecation(DBTDeprecation):
    _name = "duplicate-yaml-keys-deprecation"
    _event = "DuplicateYAMLKeysDeprecation"


class CustomTopLevelKeyDeprecation(DBTDeprecation):
    _name = "custom-top-level-key-deprecation"
    _event = "CustomTopLevelKeyDeprecation"


class CustomKeyInConfigDeprecation(DBTDeprecation):
    _name = "custom-key-in-config-deprecation"
    _event = "CustomKeyInConfigDeprecation"


class CustomKeyInObjectDeprecation(DBTDeprecation):
    _name = "custom-key-in-object-deprecation"
    _event = "CustomKeyInObjectDeprecation"


class WEOInlcudeExcludeDeprecation(DBTDeprecation):
    _name = "weo-include-exclude-deprecation"
    _event = "WEOIncludeExcludeDeprecation"


class CustomOutputPathInSourceFreshnessDeprecation(DBTDeprecation):
    _name = "custom-output-path-in-source-freshness-deprecation"
    _event = "CustomOutputPathInSourceFreshnessDeprecation"


class SourceOverrideDeprecation(DBTDeprecation):
    _name = "source-override-deprecation"
    _event = "SourceOverrideDeprecation"


class PropertyMovedToConfigDeprecation(DBTDeprecation):
    _name = "property-moved-to-config-deprecation"
    _event = "PropertyMovedToConfigDeprecation"


class ModelParamUsageDeprecation(DBTDeprecation):
    _name = "model-param-usage-deprecation"
    _event = "ModelParamUsageDeprecation"


class EnvironmentVariableNamespaceDeprecation(DBTDeprecation):
    _name = "environment-variable-namespace-deprecation"
    _event = "EnvironmentVariableNamespaceDeprecation"


class MissingPlusPrefixDeprecation(DBTDeprecation):
    _name = "missing-plus-prefix-in-config-deprecation"
    _event = "MissingPlusPrefixDeprecation"


class ArgumentsPropertyInGenericTestDeprecation(DBTDeprecation):
    _name = "arguments-property-in-generic-test-deprecation"
    _event = "ArgumentsPropertyInGenericTestDeprecation"


class MissingArgumentsPropertyInGenericTestDeprecation(DBTDeprecation):
    _name = "missing-arguments-property-in-generic-test-deprecation"
    _event = "MissingArgumentsPropertyInGenericTestDeprecation"


class ModulesItertoolsUsageDeprecation(DBTDeprecation):
    _name = "modules-itertools-usage-deprecation"
    _event = "ModulesItertoolsUsageDeprecation"


def renamed_env_var(old_name: str, new_name: str):
    class EnvironmentVariableRenamed(DBTDeprecation):
        _name = f"environment-variable-renamed:{old_name}"
        _event = "EnvironmentVariableRenamed"

    dep = EnvironmentVariableRenamed()
    deprecations_list.append(dep)
    deprecations[dep.name] = dep

    def cb():
        dep.show(old_name=old_name, new_name=new_name)

    return cb


def warn(name: str, *args, **kwargs) -> None:
    if name not in deprecations:
        # this should (hopefully) never happen
        raise RuntimeError("Error showing deprecation warning: {}".format(name))

    deprecations[name].show(*args, **kwargs)


def buffer(name: str, *args, **kwargs):
    def show_callback():
        deprecations[name].show(*args, **kwargs)

    buffered_deprecations.append(show_callback)


def show_deprecations_summary() -> None:
    summaries: List[Dict[str, Any]] = []
    for deprecation, occurrences in active_deprecations.items():
        deprecation_event = deprecations[deprecation].event()
        summaries.append(
            DeprecationSummary(
                event_name=deprecation_event.__name__,
                event_code=deprecation_event.code(),
                occurrences=occurrences,
            ).to_msg_dict()
        )

    if len(summaries) > 0:
        show_all_hint = not get_flags().show_all_deprecations
        warn_or_error(
            core_types.DeprecationsSummary(summaries=summaries, show_all_hint=show_all_hint)
        )


# these are globally available
# since modules are only imported once, active_deprecations is a singleton

active_deprecations: DefaultDict[str, int] = defaultdict(int)

deprecations_list: List[DBTDeprecation] = [
    PackageRedirectDeprecation(),
    PackageInstallPathDeprecation(),
    ConfigSourcePathDeprecation(),
    ConfigDataPathDeprecation(),
    ExposureNameDeprecation(),
    ConfigLogPathDeprecation(),
    ConfigTargetPathDeprecation(),
    CollectFreshnessReturnSignature(),
    ProjectFlagsMovedDeprecation(),
    PackageMaterializationOverrideDeprecation(),
    ResourceNamesWithSpacesDeprecation(),
    SourceFreshnessProjectHooksNotRun(),
    MFTimespineWithoutYamlConfigurationDeprecation(),
    MFCumulativeTypeParamsDeprecation(),
    MicrobatchMacroOutsideOfBatchesDeprecation(),
    GenericJSONSchemaValidationDeprecation(),
    UnexpectedJinjaBlockDeprecation(),
    DuplicateYAMLKeysDeprecation(),
    CustomTopLevelKeyDeprecation(),
    CustomKeyInConfigDeprecation(),
    CustomKeyInObjectDeprecation(),
    CustomOutputPathInSourceFreshnessDeprecation(),
    PropertyMovedToConfigDeprecation(),
    ModelParamUsageDeprecation(),
    WEOInlcudeExcludeDeprecation(),
    SourceOverrideDeprecation(),
    EnvironmentVariableNamespaceDeprecation(),
    MissingPlusPrefixDeprecation(),
    ArgumentsPropertyInGenericTestDeprecation(),
    MissingArgumentsPropertyInGenericTestDeprecation(),
    ModulesItertoolsUsageDeprecation(),
]

deprecations: Dict[str, DBTDeprecation] = {d.name: d for d in deprecations_list}

buffered_deprecations: List[Callable] = []


def reset_deprecations():
    active_deprecations.clear()


def fire_buffered_deprecations():
    [dep_fn() for dep_fn in buffered_deprecations]
    buffered_deprecations.clear()


@dataclass
class DeprecationSummary(dbtClassMixin):
    event_name: str
    event_code: str
    occurrences: int

    def to_msg_dict(self) -> Dict[str, Any]:
        return {
            "event_name": self.event_name,
            "event_code": self.event_code,
            "occurrences": self.occurrences,
        }



from dataclasses import dataclass, field
from typing import Any, Dict, Literal, Optional

from dbt.artifacts.resources.base import BaseResource
from dbt.artifacts.resources.types import NodeType
from dbt.artifacts.resources.v1.owner import Owner
from dbt_common.contracts.config.base import BaseConfig, MergeBehavior


@dataclass
class GroupConfig(BaseConfig):
    meta: Dict[str, Any] = field(default_factory=dict, metadata=MergeBehavior.Update.meta())


@dataclass
class Group(BaseResource):
    name: str
    owner: Owner
    resource_type: Literal[NodeType.Group]
    description: Optional[str] = None
    config: GroupConfig = field(default_factory=GroupConfig)



from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import List, Optional

from dbt.artifacts.resources import NodeVersion
from dbt.node_types import AccessType, NodeType


@dataclass
class ModelNodeArgs:
    name: str
    package_name: str
    identifier: str
    schema: str
    database: Optional[str] = None
    relation_name: Optional[str] = None
    version: Optional[NodeVersion] = None
    latest_version: Optional[NodeVersion] = None
    deprecation_date: Optional[datetime] = None
    access: Optional[str] = AccessType.Protected.value
    generated_at: datetime = field(
        default_factory=lambda: datetime.now(timezone.utc).replace(tzinfo=None)
    )
    depends_on_nodes: List[str] = field(default_factory=list)
    enabled: bool = True

    @property
    def unique_id(self) -> str:
        unique_id = f"{NodeType.Model}.{self.package_name}.{self.name}"
        if self.version:
            unique_id = f"{unique_id}.v{self.version}"

        return unique_id

    @property
    def fqn(self) -> List[str]:
        fqn = [self.package_name, self.name]
        # Test for None explicitly because version can be 0
        if self.version is not None:
            fqn.append(f"v{self.version}")

        return fqn



import os
import threading
import time
import traceback
from abc import ABCMeta, abstractmethod
from contextlib import nullcontext
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, List, Optional, Set

import dbt.exceptions
import dbt_common.exceptions.base
from dbt import tracking
from dbt.artifacts.resources.types import NodeType
from dbt.artifacts.schemas.results import (
    NodeStatus,
    RunningStatus,
    RunStatus,
    TimingInfo,
    collect_timing_info,
)
from dbt.artifacts.schemas.run import RunResult
from dbt.cli.flags import Flags
from dbt.compilation import Compiler
from dbt.config import RuntimeConfig
from dbt.config.profile import read_profile
from dbt.constants import DBT_PROJECT_FILE_NAME
from dbt.contracts.graph.manifest import Manifest
from dbt.events.types import (
    CatchableExceptionOnRun,
    GenericExceptionOnRun,
    InternalErrorOnRun,
    LogDbtProfileError,
    LogDbtProjectError,
    LogDebugStackTrace,
    LogSkipBecauseError,
    NodeCompiling,
    NodeConnectionReleaseError,
    NodeExecuting,
    SkippingDetails,
)
from dbt.flags import get_flags
from dbt.graph import Graph
from dbt.task import group_lookup
from dbt.task.printer import print_run_result_error
from dbt_common.events.contextvars import get_node_info
from dbt_common.events.functions import fire_event
from dbt_common.exceptions import DbtInternalError, DbtRuntimeError, NotImplementedError


def read_profiles(profiles_dir: Optional[str] = None) -> Dict[str, Any]:
    """This is only used for some error handling"""
    if profiles_dir is None:
        profiles_dir = get_flags().PROFILES_DIR

    raw_profiles = read_profile(profiles_dir)

    if raw_profiles is None:
        profiles = {}
    else:
        profiles = {k: v for (k, v) in raw_profiles.items() if k != "config"}

    return profiles


class BaseTask(metaclass=ABCMeta):
    def __init__(self, args: Flags) -> None:
        self.args = args

    def __enter__(self):
        self.orig_dir = os.getcwd()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        os.chdir(self.orig_dir)

    @abstractmethod
    def run(self):
        raise dbt_common.exceptions.base.NotImplementedError("Not Implemented")

    def interpret_results(self, results):
        return True


def get_nearest_project_dir(project_dir: Optional[str]) -> Path:
    # If the user provides an explicit project directory, use that
    # but don't look at parent directories.
    if project_dir:
        cur_dir = Path(project_dir)
        project_file = Path(project_dir) / DBT_PROJECT_FILE_NAME
        if project_file.is_file():
            return cur_dir
        else:
            raise dbt_common.exceptions.DbtRuntimeError(
                "fatal: Invalid --project-dir flag. Not a dbt project. "
                "Missing dbt_project.yml file"
            )

    cur_dir = Path.cwd()
    project_file = cur_dir / DBT_PROJECT_FILE_NAME
    if project_file.is_file():
        return cur_dir
    else:
        raise dbt_common.exceptions.DbtRuntimeError(
            "fatal: Not a dbt project (or any of the parent directories). "
            "Missing dbt_project.yml file"
        )


def move_to_nearest_project_dir(project_dir: Optional[str]) -> Path:
    nearest_project_dir = get_nearest_project_dir(project_dir)
    os.chdir(nearest_project_dir)
    return nearest_project_dir


# TODO: look into deprecating this class in favor of several small functions that
# produce the same behavior. currently this class only contains manifest compilation,
# holding a manifest, and moving direcories.
class ConfiguredTask(BaseTask):
    def __init__(
        self, args: Flags, config: RuntimeConfig, manifest: Optional[Manifest] = None
    ) -> None:
        super().__init__(args)
        self.config = config
        self.graph: Optional[Graph] = None
        self.manifest = manifest
        self.compiler = Compiler(self.config)

    def compile_manifest(self) -> None:
        if self.manifest is None:
            raise DbtInternalError("compile_manifest called before manifest was loaded")

        start_compile_manifest = time.perf_counter()

        self.graph = self.compiler.compile(self.manifest)

        compile_time = time.perf_counter() - start_compile_manifest
        if dbt.tracking.active_user is not None:
            dbt.tracking.track_runnable_timing({"graph_compilation_elapsed": compile_time})

    @classmethod
    def from_args(cls, args: Flags, *pargs, **kwargs):
        move_to_nearest_project_dir(args.project_dir)
        try:
            # This is usually RuntimeConfig
            config = RuntimeConfig.from_args(args)
        except dbt.exceptions.DbtProjectError as exc:
            fire_event(LogDbtProjectError(exc=str(exc)))

            tracking.track_invalid_invocation(args=args, result_type=exc.result_type)
            raise dbt_common.exceptions.DbtRuntimeError("Could not run dbt") from exc
        except dbt.exceptions.DbtProfileError as exc:
            all_profile_names = list(read_profiles(get_flags().PROFILES_DIR).keys())
            fire_event(LogDbtProfileError(exc=str(exc), profiles=all_profile_names))
            tracking.track_invalid_invocation(args=args, result_type=exc.result_type)
            raise dbt_common.exceptions.DbtRuntimeError("Could not run dbt") from exc
        return cls(args, config, *pargs, **kwargs)


class ExecutionContext:
    """During execution and error handling, dbt makes use of mutable state:
    timing information and the newest (compiled vs executed) form of the node.
    """

    def __init__(self, node) -> None:
        self.timing: List[TimingInfo] = []
        self.node = node


class BaseRunner(metaclass=ABCMeta):
    def __init__(self, config, adapter, node, node_index: int, num_nodes: int) -> None:
        self.config = config
        self.compiler = Compiler(config)
        self.adapter = adapter
        self.node = node
        self.node_index = node_index
        self.num_nodes = num_nodes

        self.skip = False
        self.skip_cause: Optional[RunResult] = None

        self.run_ephemeral_models = False

    @abstractmethod
    def compile(self, manifest: Manifest) -> Any:
        pass

    def _node_build_path(self) -> Optional[str]:
        return self.node.build_path if hasattr(self.node, "build_path") else None

    def get_result_status(self, result) -> Dict[str, str]:
        if result.status == NodeStatus.Error:
            return {"node_status": "error", "node_error": str(result.message)}
        elif result.status == NodeStatus.Skipped:
            return {"node_status": "skipped"}
        elif result.status == NodeStatus.Fail:
            return {"node_status": "failed"}
        elif result.status == NodeStatus.Warn:
            return {"node_status": "warn"}
        else:
            return {"node_status": "passed"}

    def run_with_hooks(self, manifest):
        if self.skip:
            return self.on_skip()

        # no before/after printing for ephemeral mdoels
        if not self.node.is_ephemeral_model:
            self.before_execute()

        result = self.safe_run(manifest)
        self.node.update_event_status(
            node_status=result.status,
            finished_at=datetime.now(timezone.utc).replace(tzinfo=None).isoformat(),
        )

        if not self.node.is_ephemeral_model:
            self.after_execute(result)

        return result

    def _build_run_result(
        self,
        node,
        start_time,
        status,
        timing_info,
        message,
        agate_table=None,
        adapter_response=None,
        failures=None,
        batch_results=None,
    ):
        execution_time = time.time() - start_time
        thread_id = threading.current_thread().name
        if adapter_response is None:
            adapter_response = {}
        return RunResult(
            status=status,
            thread_id=thread_id,
            execution_time=execution_time,
            timing=timing_info,
            message=message,
            node=node,
            agate_table=agate_table,
            adapter_response=adapter_response,
            failures=failures,
            batch_results=batch_results,
        )

    def error_result(self, node, message, start_time, timing_info):
        return self._build_run_result(
            node=node,
            start_time=start_time,
            status=RunStatus.Error,
            timing_info=timing_info,
            message=message,
        )

    def ephemeral_result(self, node, start_time, timing_info):
        return self._build_run_result(
            node=node,
            start_time=start_time,
            status=RunStatus.Success,
            timing_info=timing_info,
            message=None,
        )

    def from_run_result(self, result, start_time, timing_info):
        return self._build_run_result(
            node=result.node,
            start_time=start_time,
            status=result.status,
            timing_info=timing_info,
            message=result.message,
            agate_table=result.agate_table,
            adapter_response=result.adapter_response,
            failures=result.failures,
            batch_results=result.batch_results,
        )

    def compile_and_execute(self, manifest: Manifest, ctx: ExecutionContext):
        result = None
        with (
            self.adapter.connection_named(self.node.unique_id, self.node)
            if get_flags().INTROSPECT
            else nullcontext()
        ):
            ctx.node.update_event_status(node_status=RunningStatus.Compiling)
            fire_event(
                NodeCompiling(
                    node_info=ctx.node.node_info,
                )
            )
            with collect_timing_info("compile", ctx.timing.append):
                # if we fail here, we still have a compiled node to return
                # this has the benefit of showing a build path for the errant
                # model.  This calls the 'compile' method in CompileTask
                ctx.node = self.compile(manifest)

            # for ephemeral nodes, we only want to compile, not run
            if not ctx.node.is_ephemeral_model or self.run_ephemeral_models:
                ctx.node.update_event_status(node_status=RunningStatus.Executing)
                fire_event(
                    NodeExecuting(
                        node_info=ctx.node.node_info,
                    )
                )
                with collect_timing_info("execute", ctx.timing.append):
                    result = self.run(ctx.node, manifest)
                    ctx.node = result.node

        return result

    def _handle_catchable_exception(self, e: DbtRuntimeError, ctx: ExecutionContext) -> str:
        if e.node is None:
            e.add_node(ctx.node)

        fire_event(
            CatchableExceptionOnRun(
                exc=str(e), exc_info=traceback.format_exc(), node_info=get_node_info()
            )
        )
        return str(e)

    def _handle_internal_exception(self, e: DbtInternalError, ctx: ExecutionContext) -> str:
        fire_event(
            InternalErrorOnRun(
                build_path=self._node_build_path(), exc=str(e), node_info=get_node_info()
            )
        )
        return str(e)

    def _handle_generic_exception(self, e: Exception, ctx: ExecutionContext) -> str:
        fire_event(
            GenericExceptionOnRun(
                build_path=self._node_build_path(),
                unique_id=self.node.unique_id,
                exc=str(e),
                node_info=get_node_info(),
            )
        )
        fire_event(LogDebugStackTrace(exc_info=traceback.format_exc()))

        return str(e)

    def handle_exception(self, e: Exception, ctx: ExecutionContext) -> str:
        if isinstance(e, DbtRuntimeError):
            error = self._handle_catchable_exception(e, ctx)
        elif isinstance(e, DbtInternalError):
            error = self._handle_internal_exception(e, ctx)
        else:
            error = self._handle_generic_exception(e, ctx)
        return error

    def safe_run(self, manifest: Manifest):
        started = time.time()
        ctx = ExecutionContext(self.node)
        error = None
        result = None

        try:
            result = self.compile_and_execute(manifest, ctx)
        except Exception as e:
            error = self.handle_exception(e, ctx)
        finally:
            exc_str = self._safe_release_connection()

            # if releasing failed and the result doesn't have an error yet, set
            # an error
            if (
                exc_str is not None
                and result is not None
                and result.status != NodeStatus.Error
                and error is None
            ):
                error = exc_str

        if error is not None:
            result = self.error_result(ctx.node, error, started, ctx.timing)
        elif result is not None:
            result = self.from_run_result(result, started, ctx.timing)
        else:
            result = self.ephemeral_result(ctx.node, started, ctx.timing)
        return result

    def _safe_release_connection(self):
        """Try to release a connection. If an exception is hit, log and return
        the error string.
        """
        try:
            self.adapter.release_connection()
        except Exception as exc:
            fire_event(
                NodeConnectionReleaseError(
                    node_name=self.node.name, exc=str(exc), exc_info=traceback.format_exc()
                )
            )
            return str(exc)

        return None

    def before_execute(self) -> None:
        raise NotImplementedError("before_execute is not implemented")

    def execute(self, compiled_node, manifest):
        raise NotImplementedError("execute is not implemented")

    def run(self, compiled_node, manifest):
        return self.execute(compiled_node, manifest)

    def after_execute(self, result) -> None:
        raise NotImplementedError("after_execute is not implemented")

    def _skip_caused_by_ephemeral_failure(self) -> bool:
        if self.skip_cause is None or self.skip_cause.node is None:
            return False
        return self.skip_cause.node.is_ephemeral_model

    def on_skip(self):
        schema_name = getattr(self.node, "schema", "")
        node_name = self.node.name

        error_message = None
        if not self.node.is_ephemeral_model:
            # if this model was skipped due to an upstream ephemeral model
            # failure, print a special 'error skip' message.
            # Include skip_cause NodeStatus
            group = group_lookup.get(self.node.unique_id)

            if self._skip_caused_by_ephemeral_failure():
                fire_event(
                    LogSkipBecauseError(
                        schema=schema_name,
                        relation=node_name,
                        index=self.node_index,
                        total=self.num_nodes,
                        status=self.skip_cause.status,
                        group=group,
                    )
                )
                # skip_cause here should be the run_result from the ephemeral model
                print_run_result_error(result=self.skip_cause, newline=False)
                if self.skip_cause is None:  # mypy appeasement
                    raise DbtInternalError(
                        "Skip cause not set but skip was somehow caused by an ephemeral failure"
                    )
                # set an error so dbt will exit with an error code
                error_message = (
                    "Compilation Error in {}, caused by compilation error "
                    "in referenced ephemeral model {}".format(
                        self.node.unique_id, self.skip_cause.node.unique_id
                    )
                )
            else:
                # 'skipped' nodes should not have a value for 'node_finished_at'
                # they do have 'node_started_at', which is set in GraphRunnableTask.call_runner
                self.node.update_event_status(node_status=RunStatus.Skipped)
                fire_event(
                    SkippingDetails(
                        resource_type=self.node.resource_type,
                        schema=schema_name,
                        node_name=node_name,
                        index=self.node_index,
                        total=self.num_nodes,
                        node_info=self.node.node_info,
                        group=group,
                    )
                )

        node_result = RunResult.from_node(self.node, RunStatus.Skipped, error_message)
        return node_result

    def do_skip(self, cause=None) -> None:
        self.skip = True
        self.skip_cause = cause


def resource_types_from_args(
    args: Flags, all_resource_values: Set[NodeType], default_resource_values: Set[NodeType]
) -> Set[NodeType]:

    if not args.resource_types:
        resource_types = default_resource_values
    else:
        # This is a list of strings, not NodeTypes
        arg_resource_types = set(args.resource_types)

        if "all" in arg_resource_types:
            arg_resource_types.remove("all")
            arg_resource_types.update(all_resource_values)
        if "default" in arg_resource_types:
            arg_resource_types.remove("default")
            arg_resource_types.update(default_resource_values)
        # Convert to a set of NodeTypes now that the non-NodeType strings are gone
        resource_types = set([NodeType(rt) for rt in list(arg_resource_types)])

    if args.exclude_resource_types:
        # Convert from a list of strings to a set of NodeTypes
        exclude_resource_types = set([NodeType(rt) for rt in args.exclude_resource_types])
        resource_types = resource_types - exclude_resource_types

    return resource_types
